<?php
/* 
=================================================================================
 + 파일명 : multiSendMail.php
 + 버전: 1.0
 + 배포일 : 2022-08-21
 + 제작자 : R BLOG (lcy@redinfo.co.kr)
 + 기타   : (가이드 URL) https://blog.redinfo.co.kr/post/view?pid=116
            (블로그 URL) https://blog.redinfo.co.kr
=================================================================================
*/
namespace multiSendMail;
class multiSendMail{

    /* 기본설정 */
    private $config = array(
        'module'=>'smtp', // 발송타입 (smtp / phpmail / auto)

        /*SMTP 설정 필수*/
        'smtp.host'=>'',  // SMTP 호스트
        'smtp.port'=>25, // SMTP 포트
        'smtp.username'=>'', // SMTP ID
        'smtp.password'=>'', // SMTP PW
        'smtp.socket_timout'=>10, // 소켓연결 타임아웃값

        /*공통 설정*/
        'debug'=>0, // 디버그 - 0: 미사용, 1: 사용
        'msg'=>0, // 메시징뷰 - 0: 미사용, 1: 사용
        'charset'=>'UTF-8', // 언어셋
        'ctype'=>'text/html', // 내용 컨텐츠타입

        /* 비동기 소켓설정 (서버에 따라 지원안될 수 있음)  */
        'async.use'=> 0, // 비동기 소켓 사용여부 - 0: 미사용, 1: 사용
        'async.url'=>'', // 비동기 소켓 url 
        'async.key'=>'multiSendMail', // 비동기 소켓 통신 키값 - 필수 
        'async.encrypt'=>'md5', // 비동기 소켓 통신 키값 암호화 방식, 미 지정시 md5 로 지정
    );

    /* 기본설정 백업 */
    private $clearConfig = array();

    /*기본옵션*/
    private $option = array(    
        'to'=> array(), // 받는이
        'from'=>array(), // 보내는 이 
        'subject'=>'', // 메일 제목
        'body'=>'', // 메일 내용 
        'cc'=>array(), // 참조
        'bcc'=>array(), // 숨은 참조
    );   

    /* 아래  변수는 수정 금지 */
    private $isAsyncSocketClient = false; // 비동기 소켓통신 설정변수
    private $parts = array(); // 첨부파일 관련 변수
    private $module; // auto 일경우 자동 모듈 지정
    private $fp; // socket 
    private $length = 4096; // socket length
    private $connection = false;
    private $msg  = array(); // 메시지, 첨부파일 스택 배열
    private $newline = "\r\n"; // 뉴라인(표준화)
    private $error = array(
        'line' => 0,
        'msg' => NULL,
    );

    function __construct($config = array()){

    	$this->clearConfig  = $this->config; // 설정 백업
        foreach($this->config as $k=>$v){ if( !empty($config[$k])){$this->config[$k] =  $config[$k]; }}
    }

    function __destruct(){ 
        $this->msg('multiSendMail 모듈 종료',__LINE__);
    }


	/* 비동기 소켓통신으로 전달받았는지 체크 */    
    private function isAsyncSocketClient(){
    	$config = $this->config;
    	$post = $_POST;
    	if(empty($config['async.encrypt'])){ $config['async.encrypt'] = 'md5'; }
    	if( empty($post['AUTH_ASYNC_SOCKET_CLIENT_KEY'])){ return false; }

    	// AUTH_ASYNC_SOCKET_CLIENT_KEY 값이 있을 경우 비동기 통신으로 접근한것으로 간주.
    	$this->isAsyncSocketClient = true;

    	$checkKey = hash($config['async.encrypt'],$config['async.key']);
    	$hashKey = $post['AUTH_ASYNC_SOCKET_CLIENT_KEY'];
    	if($hashKey != $checkKey){
    		return  $this->debug("비동기 소켓통신 응답 - 실패(키값 불일치 - key[".$hashKey."] ) ",__LINE__); 
    	}
    	if( empty($post[$hashKey])){
    		return  $this->debug("비동기 소켓통신 응답 - 실패(POST 데이터 손실) ",__LINE__); 	
    	}
    	$datas = unserialize(base64_decode($post[$hashKey]));

    	if( !isset($datas['config']) || !isset($datas['option'])){
    		return  $this->debug("비동기 소켓통신 응답 - 실패(설정값 또는 옵션값 누락) ",__LINE__); 
    	}

    	if( !isset($datas['parts']) ){ $datas['parts'] = array();  }
    	if( count($datas['parts']) > 0){
    		foreach($datas['parts'] as $k=>$v){   			
    			$this->setAttach($v['path'],$v['name'],$v['ctype']);
    		}
    	}
    	$this->config = $datas['config'];
    	$this->option = $datas['option'];

    	$this->msg('비동기 소켓통신 응답 - 완료',__LINE__);

    	return true;
    }

    /* 비동기 소켓통신 함수  */
    private function applyAsyncSocketClient($url = '', $params = array()){
    	$config = $this->config; 
		$post_params = array();

		$post_string = http_build_query($params);  
		$parts=parse_url($url);  

		if ($parts['scheme'] == 'http')  
		{  
		    $fp = fsockopen($parts['host'], isset($parts['port'])?$parts['port']:80, $errno, $errstr, $config['smtp.socket_timout']);  
		}  
		else if ($parts['scheme'] == 'https')  
		{  
		    $fp = fsockopen("ssl://" . $parts['host'], isset($parts['port'])?$parts['port']:443, $errno, $errstr, $config['smtp.socket_timout']);  
		}  

		$out = "POST ".$parts['path']." HTTP/1.1".$this->newline;  
		$out.= "Host: ".$parts['host'].$this->newline;  
		$out.= "Content-Type: application/x-www-form-urlencoded".$this->newline;  
		$out.= "Content-Length: ".strlen($post_string).$this->newline;  
		$out.= "Connection: Close".$this->newline.$this->newline;  
		$out.= $post_string;  
		fwrite($fp, $out);  
		fclose($fp);  
    }

    /* php mail 전송방식 */
    private function phpmail(){
        $args = func_get_args();
        if( count($args ) > 0){
            $method = $args[0];
            unset($args[0]);
            $args = array_values($args);
            if(!empty($method)) $method = strtolower($method);
            else{ $method = null; }
        }else{ $args  = array(); }
        if( $method === null ){ return false;}     

        switch($method){

            case "mailer":

            	$this->msg('PHPMAIL - 전송 준비',__LINE__);

            	$option = $this->option;

            	// 옵션을 체크 
		        if( empty($option['to']) || !is_array($option['to']) || count($option['to']) < 1 ){ return $this->debug("`옵션: to` 데이터가 누락되었습니다.",__LINE__);  }
		        if( empty($option['from']) || !is_array($option['from']) || count($option['from']) < 1 ){ return $this->debug("`옵션: from` 데이터가 누락되었습니다.",__LINE__);  }
		        if( empty($option['subject'])){ return $this->debug("`옵션: subject` 데이터가 누락되었습니다.",__LINE__);  }
		        if( empty($option['body'])){ return $this->debug("`옵션: body` 데이터가 누락되었습니다.",__LINE__);  }

		        if( !is_array($option['cc'])){ $option['cc'] = array(); }
		        if( !is_array($option['bcc'])){ $option['bcc'] = array(); }

		        // 메일 보낸다. 
		        $success = 0;
	        	$data = array();
	        	$data['from'] = $option['from']; // 보내는이/반송
	        	$data['to'] = $option['to'];
	        	$data['subject'] = $option['subject'];
	        	$data['body'] = $option['body'];
	        	$data['cc'] = $option['cc']; // 참조
	        	$data['bcc'] = $option['bcc']; // 숨은참조
	        
	        	$request = $this->phpmail('send',$data);
	        	$this->close(); 

	        	if($request === true){
	        		$this->msg('PHPMAIL - 메일 전송 성공',__LINE__);
	        	}else{
					$this->msg('PHPMAIL - 메일 전송 실패',__LINE__);	        		
	        	}

	        	return $request;

            break;

            case "send":
                $parmSet = $this->support('parm',$args,array('data'=>array()));
                extract($parmSet);
                $config = $this->config;

            	// 보내는이 메일 체크 
            	if( $this->support('emailChk',$data['from']['email']) !== true){ return $this->debug(__LINE__*-1,"`from` 이메일 형식 오류(".$data['from']['email']." 은 잘못된 이메일 입니다.)"); }
		        $name = !empty($data['from']['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($data['from']['name']).'?=' : null;
		        $subject = '=?'.$config['charset'].'?B?'.base64_encode($data['subject']).'?=';

				// 컨텐츠 검증
				$contentType = $config['ctype'];
				
				$this->msg('PHPMAIL - 발송데이터 설정 준비',__LINE__);
		        
				// Header
				$header = "From: ".(empty($name) ? null : $name)."<".$data['from']['email'].">".$this->newline;
				$header .= "Reply-To: ".$data['from']['email'].$this->newline;

				// 파일데이터
				$parts = array_filter($this->parts);
				$attcnt = count($parts);

				// boundary 데이터 
				$boundary = $this->support('boundary');

				if( $attcnt > 0){
					$header .= "Content-Type: multipart/mixed; boundary=\"PHP-mixed-".$boundary."\""; 
				}else{
					$header .= "Content-Type: ".$contentType."; charset=".$config['charset']."".$this->newline;
				}

				$this->msg('PHPMAIL - 헤더 설정 완료',__LINE__);

				// 받는이 데이터 가공 
            	$arrMimeTo = array();
            	foreach($data['to'] as $to){
            		if( $this->support('emailChk',$to['email']) !== true){
            			return $this->debug(__LINE__*-1,"`to` 이메일 형식 오류(".$to['email']." 은 잘못된 이메일 입니다.)");
            		}	
			        $name = !empty($to['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($to['name']).'?=' : null;
			        $arrMimeTo[] =$name."<".$to['email'].">";		        
            	}
            	$to = implode(",",$arrMimeTo);

            	$this->msg('PHPMAIL - 받는이 설정 완료('.$to.')',__LINE__);

            	// 참조 데이터 가공
            	$arrMimeTo = array();
            	if( count($data['cc']) > 0){
	            	foreach($data['cc'] as $to){
	            		if( $this->support('emailChk',$to['email']) !== true){
	            			return $this->debug(__LINE__*-1,"`cc` 이메일 형식 오류(".$to['email']." 은 잘못된 이메일 입니다.)");
	            		}	

				        $name = !empty($to['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($to['name']).'?=' : null;
				        $arrMimeTo[] =$name."<".$to['email'].">";		        
	            	}
	            	if( count($arrMimeTo) > 0){
	            		$cc = implode(",",$arrMimeTo);
	            		$header .= "Cc: ".$cc.".\r\n";
	            		$this->msg('PHPMAIL - 참조 설정 완료('.$cc.')',__LINE__);
	            	}
            	}

            	// 숨은참조 데이터 가공
            	$arrMimeTo = array();
            	if( count($data['bcc']) > 0){
	            	foreach($data['bcc'] as $to){
	            		if( $this->support('emailChk',$to['email']) !== true){
	            			return $this->debug(__LINE__*-1,"`bcc` 이메일 형식 오류(".$to['email']." 은 잘못된 이메일 입니다.)");
	            		}	
				        $name = !empty($to['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($to['name']).'?=' : null;
				        $arrMimeTo[] =$name."<".$to['email'].">";		        
	            	}
	            	if( count($arrMimeTo) > 0){
	            		$bcc = implode(",",$arrMimeTo);
	            		$header .= "Bcc: ".$bcc.".\r\n";
	            		$this->msg('PHPMAIL - 숨은참조 설정 완료'.$bcc.'',__LINE__);
	            	}
            	}

            	if( $attcnt > 0){

            		$this->msg('PHPMAIL - 파일첨부 데이터 설정 준비',__LINE__);

					// Body
					$body = "--PHP-mixed-".$boundary.$this->newline;
					$body .= "Content-Type: multipart/alternative; boundary=\"PHP-alt-".$boundary."\"".$this->newline.$this->newline;
					$body .= "--PHP-alt-".$boundary.$this->newline;
					$body .= "Content-Type: ".$contentType."; charset=\"".$config['charset']."\"".$this->newline;
					$body .= "Content-Transfer-Encoding: base64".$this->newline.$this->newline;
					$body .= chunk_split(base64_encode($data['body']));
					$body .= $this->newline.$this->newline;
				
					// MESSAGE - END
					$body .= "--PHP-alt-".$boundary."-- ".$this->newline;

					// FILE ATTACHMENT
					if( count($parts) > 0) {
						foreach($parts as $part){

							// 파일이름 가공
							if(!empty($part['name'])){
								$filename = '=?'.$config['charset'].'?B?'.base64_encode($part['name']).'?=';
							}else{ continue; }

							// 파일가공
							$attachment = chunk_split(base64_encode($part['message'])); 

							$body .= "--PHP-mixed-".$boundary."".$this->newline;
							$body .= "Content-Type: ".$part['ctype'].";";
							if(!empty($part['name'])) $body .= " name=\"".$filename."\"".$this->newline;
							$body .= "Content-Transfer-Encoding: base64".$this->newline;
							$body .= "Content-Disposition: attachment; filename=\"".$filename."\"".$this->newline.$this->newline;
							$body .= $attachment."".$this->newline;
						}
						$body .= "--PHP-mixed-".$boundary."--".$this->newline;
					}

					$this->msg('PHPMAIL - 파일첨부 데이터 설정 완료(총 '.number_format($attcnt).'개)',__LINE__);
            	}
            	// 첨부파일 없을 시 일반전송
            	else{
            		$body = $data['body'];
            		$this->msg('PHPMAIL - 일반 데이터 설정 완료',__LINE__);
            	}

            	$send = mail( $to, $subject, $body, $header );
				return $send;
            break;

        }    	
    }

    /* smtp 전송방식 */
    private function smtp(){	
        $args = func_get_args();
        if( count($args ) > 0){
            $method = $args[0];
            unset($args[0]);
            $args = array_values($args);
            if(!empty($method)) $method = strtolower($method);
            else{ $method = null; }
        }else{ $args  = array(); }
        if( $method === null ){ return false;}     

        switch($method){
            case "connect":
            	$config = $this->config;

            	$this->msg('SMTP - 전송 준비',__LINE__);

		        if( empty($config['smtp.host']) ){ return $this->debug(" `host`가 누락되었습니다.",__LINE__); }
		        if( empty($config['smtp.port']) ){ return $this->debug(" `port`가 누락되었습니다.",__LINE__); }
		        if( $config['smtp.port'] == 465 && !preg_match("/^ssl:\/\//",$config['smtp.host'])){ $config['smtp.host'] = "ssl://".$config['smtp.host'];}
		        $this->msg("SMTP - 연결 ( HOST - ".$config['smtp.host']." , PORT - ".$config['smtp.port'].") Connecting...",__LINE__);
  				

		        if( function_exists('stream_socket_client')){
		        	$context = stream_context_create();
		        	$this->fp = stream_socket_client(gethostbyname($config['smtp.host']).":".$config['smtp.port'],$errno,$errstr,$config['smtp.socket_timout'],STREAM_CLIENT_CONNECT,$context);
		        }
		        else{
		        	$this->fp = fsockopen($config['smtp.host'], $config['smtp.port'], $errno, $errstr, $config['smtp.socket_timout']);
		        }

		        if(!$this->fp){
		        	return $this->debug("SMTP(".$config['smtp.host'].":".$config['smtp.port'].") 서버접속에 실패했습니다.[".$errno.":".$errstr."]",__LINE__);	
		        }

				
		        $this->smtp('dialogue',220); 

		        // [2022-07-04] 587: 포트 지원추가 
		        if( $config['smtp.port'] == 587){
		            if(!$this->smtp('dialogue',220, "STARTTLS")) { return $this->debug("`TLS` 지원불가능",__LINE__); }
		            if(false == stream_socket_enable_crypto($this->fp, true, STREAM_CRYPTO_METHOD_TLS_CLIENT)){
		               return $this->debug("`TLS` 지원불가능",__LINE__);
		            }       
		        }

		       	$this->smtp('dialogue',0,"EHLO ".$_SERVER['HTTP_HOST']);
		        if( ($checkPoint = $this->smtp('dialogue',334,"AUTH LOGIN")) !== true ){  return $this->debug($checkPoint,__LINE__); }
		        if( ($checkPoint = $this->smtp('dialogue',334, base64_encode($config['smtp.username']) )) !== true ){  return $this->debug($checkPoint,__LINE__); }
		        if( ($checkPoint = $this->smtp('dialogue',235, base64_encode($config['smtp.password']) )) !== true ){  return $this->debug($checkPoint,__LINE__); }

		        $this->msg('SMTP - AUTH 인증성공',__LINE__);

		        return true;
            break;

            case "mailer":
            	$option = $this->option;

            	// 옵션을 체크 
		        if( empty($option['to']) || !is_array($option['to']) || count($option['to']) < 1 ){ return $this->debug("`옵션: to` 데이터가 누락되었습니다.",__LINE__);  }
		        if( empty($option['from']) || !is_array($option['from']) || count($option['from']) < 1 ){ return $this->debug("`옵션: from` 데이터가 누락되었습니다.",__LINE__);  }
		        if( empty($option['subject'])){ return $this->debug("`옵션: subject` 데이터가 누락되었습니다.",__LINE__);  }
		        if( empty($option['body'])){ return $this->debug("`옵션: body` 데이터가 누락되었습니다.",__LINE__);  }

		        if( !is_array($option['cc'])){ $option['cc'] = array(); }
		        if( !is_array($option['bcc'])){ $option['bcc'] = array(); }


                // SMTP 연결한다.
                if(!$this->connection){ $this->connection = $this->smtp('connect');  }
            	if(!$this->connection){  return false;   }

		        // 메일의 데이터를 생성한다.
		        $mimedata = $this->smtp('mimedata');

		        // 메일 보낸다. 
		        $success = 0;
	        	$data = array();
	        	$data['from'] = $option['from']; // 보내는이/반송
	        	$data['to'] = $option['to'];
	        	$data['mimedata'] = $mimedata; // 마임데이터
	        	$data['cc'] = $option['cc']; // 참조
	        	$data['bcc'] = $option['bcc']; // 숨은참조
	        
	        	$request = $this->smtp('send',$data);	
	        	$this->close();      

	        	if($request === true){
	        		$this->msg('SMTP - 메일 전송 성공',__LINE__);
	        	}

	        	return $request;
		        
            break;

            case "send":
                $parmSet = $this->support('parm',$args,array('data'=>array()));
                extract($parmSet);
                $config = $this->config;

                // 마임생성
                $mime = "Message-ID: <".$this->support('get_msgid').">".$this->newline;

            	// 보내는이 메일 체크 
            	if( $this->support('emailChk',$data['from']['email']) !== true){ return $this->debug(__LINE__*-1,"`from` 이메일 형식 오류(".$data['from']['email']." 은 잘못된 이메일 입니다.)"); }
		        if( ($checkPoint = $this->smtp('dialogue',250, "MAIL FROM:<".$data['from']['email'].">")) !== true ){
		        	return $this->debug($checkPoint,__LINE__);
		        }
		        $name = !empty($data['from']['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($data['from']['name']).'?=' : null;
		        $mime .= "From: ".(empty($name) ? null : $name)."<".$data['from']['email'].">".$this->newline;

            	// 받는이 체크/가공
            	$arrMimeTo = array();
            	foreach($data['to'] as $to){
            		if( $this->support('emailChk',$to['email']) !== true){
            			return $this->debug(__LINE__*-1,"`to` 이메일 형식 오류(".$to['email']." 은 잘못된 이메일 입니다.)");
            		}	

			        if( ($checkPoint = $this->smtp('dialogue',250, "RCPT TO:<".$to['email'].">")) !== true ){			        	
			        	return $this->debug($checkPoint,__LINE__);
			        }
			        $name = !empty($to['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($to['name']).'?=' : null;
			        $arrMimeTo[] =$name."<".$to['email'].">";		        
            	}
            	$mime .= "To: ".implode(",",$arrMimeTo).$this->newline;


            	// 참조 데이터 가공
            	$arrMimeTo = array();
            	if( count($data['cc']) > 0){
	            	foreach($data['cc'] as $to){
	            		if( $this->support('emailChk',$to['email']) !== true){
	            			return $this->debug(__LINE__*-1,"`cc` 이메일 형식 오류(".$to['email']." 은 잘못된 이메일 입니다.)");
	            		}	

				        if( ($checkPoint = $this->smtp('dialogue',250, "RCPT TO:<".$to['email'].">")) !== true ){			        	
				        	return $this->debug($checkPoint,__LINE__);
				        }

				        $name = !empty($to['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($to['name']).'?=' : null;
				        $arrMimeTo[] =$name."<".$to['email'].">";		        
	            	}
	            	if( count($arrMimeTo) > 0){
	            		$mime .= "Cc: ".implode(",",$arrMimeTo).$this->newline;
	            	}
            	}

            	// 숨은참조 데이터 가공
            	$arrMimeTo = array();
            	if( count($data['bcc']) > 0){
	            	foreach($data['bcc'] as $to){
	            		if( $this->support('emailChk',$to['email']) !== true){
	            			return $this->debug(__LINE__*-1,"`bcc` 이메일 형식 오류(".$to['email']." 은 잘못된 이메일 입니다.)");
	            		}	

				        if( ($checkPoint = $this->smtp('dialogue',250, "RCPT TO:<".$to['email'].">")) !== true ){			        	
				        	return $this->debug($checkPoint,__LINE__);
				        }

				        $name = !empty($to['name']) ? '=?'.$config['charset'].'?B?'.base64_encode($to['name']).'?=' : null;
				        $arrMimeTo[] =$name."<".$to['email'].">";		        
	            	}
	            	if( count($arrMimeTo) > 0){
	            		$mime .= "Bcc: ".implode(",",$arrMimeTo).$this->newline;
	            	}
            	}

		        $this->smtp('dialogue',354, "DATA");
		        fwrite($this->fp, $mime);
		        fwrite($this->fp, $data['mimedata']);
		        if( ($checkPoint = $this->smtp('dialogue',250, ".")) !== true ){	
		            return false;
		        }

		        return true;

            break;

            // 파일데이터
            case "mimefile":
                $parmSet = $this->support('parm',$args,array('part'=>array()));
                extract($parmSet);          	

                $config= $this->config;

		        $part = array_filter($part);
		        if( empty($part) || !is_array($part) || count($part)< 1){ return false; }

		        $name = '=?'.$config['charset'].'?B?'.base64_encode($part['name']).'?=';
		        $mimefile = "Content-Type: ".$part['ctype'];
		        if(!empty($part['name'])) $mimefile .= "; name=\"".$name."\"";
		        $mimefile .= $this->newline."Content-Transfer-Encoding: base64".$this->newline;
		        $mimefile .= "Content-Disposition: attachment; filename=\"".$name."\"".$this->newline.$this->newline;
		        $mimefile .= chunk_split(base64_encode($part['message']))."";
		        return $mimefile;            	
            break;

            // 마임데이터
            case "mimedata":            	
            	$option = $this->option;
            	$config = $this->config;
            	$parts = $this->parts;

		        $boundary = $this->support('boundary');
		        $attcnt = count($parts);

		        $subject = '=?'.$config['charset'].'?B?'.base64_encode($option['subject']).'?=';
		        $mimedata = "Subject: ".$subject.$this->newline;
		        $mimedata .= "Date: ".date ("D, j M Y H:i:s T",time()).$this->newline;
		        $mimedata .= "MIME-Version: 1.0".$this->newline;
		        if($attcnt > 0) {
		        	$this->msg('SMTP - 파일첨부 데이터 설정 준비',__LINE__);
		            $mimedata .= "Content-Type: multipart/mixed; boundary=\"".$boundary."\"".$this->newline.$this->newline;
		            $mimedata .= "This is a multi-part message in MIME format.".$this->newline.$this->newline;
		            $mimedata .= "--".$boundary.$this->newline;
		        }
		        $mimedata .= "Content-Type: ".$config['ctype']."; charset=\"".$config['charset']."\"".$this->newline;
		        $mimedata .= "Content-Transfer-Encoding: base64" .$this->newline.$this->newline. chunk_split(base64_encode($option['body']));
		            
		        // 첨부파일이 있을 경우에만 생성
		        if($attcnt > 0) {
		            $mimedata .= $this->newline.$this->newline."--".$boundary;
		            for($i=0; $i<$attcnt; $i++) {
		                $mimedata .= "\r\n".$this->smtp('mimefile',$parts[$i]).$this->newline.$this->newline."--".$boundary;
		            }
		            $mimedata .= "--".$this->newline;
		            $this->msg('SMTP - 파일첨부 데이터 설정 완료(총 '.number_format($attcnt).'개)',__LINE__);
		        }

		        return $mimedata;            	
            break;

            case "dialogue":
                $parmSet = $this->support('parm',$args,array('code'=>'','cmd'=>'','streammode'=> false));
                extract($parmSet);  
                $code =  (int)$code;

                if(!empty($cmd)){ 
                	fwrite($this->fp,$cmd.$this->newline); $this->msg('<strong style="color:95afc9;">'.htmlspecialchars($cmd).'</strong>',__LINE__);
                }

                if($code === 0){
					stream_set_timeout($this->fp,1);
					$lines = stream_get_contents($this->fp);
				}
				else{
					$lines = fread($this->fp,$this->length);
				}
				
		        if( empty($lines)){  return $this->debug("`".htmlspecialchars($cmd)."` 응답에러",__LINE__); }
		        $arrLines = array_filter(explode($this->newline,$lines));
		        $arrLines = array_values($arrLines);

		        $codeCheck = empty($code) ? true : false;
		        $checkPoint = false;
		        foreach($arrLines as $line){
		        	$this->msg($line,__LINE__);
			        if(!empty($code)){
			        	preg_match("/^([0-9]+).(.*)$/", $line, $matches);
			        	if( $matches[1] == $code){ $codeCheck = true; }
			        	else{
							return "`".htmlspecialchars($cmd)."` 에러(".$line.")";
			        	}
			        }		        	
		        }

		        return $codeCheck;
            break;
        }   
    }

    /* 설정체크 */
    private function configCheck(){
        $config = $this->config;

         // 필수 type 체크 
        if( empty($config['module']) || (isset($config['module']) && in_array($config['module'], array('smtp','phpmail','auto')) < 1)){
        	return $this->debug("`module` 설정값이 잘못되었습니다.(smtp, phpmail, auto 중 선택해주세요)",__LINE__);  
        }

         // 호스트 체크 
        $cv = array('smtp.host','smtp.port','smtp.username','smtp.password');
        $checksumSmtp = true;
        foreach($cv as $k=>$v){
            if( empty($config[$v])){ $checksumSmtp = $v; break; }
            if( $v == 'smtp.host' && gethostbyname($config[$v]) == $config[$v] ){$checksumSmtp = $v;  }
        }
        
         // smtp 체크 
        if( $config['module'] == 'smtp' && $checksumSmtp !== true){
            return $this->debug("필수설정 항목 `".$checksumSmtp."` 값이 잘못되었습니다.",__LINE__);
        }

        // 기타 항목 설정 값 체크 
       	if(empty($config['charset'])){  return $this->debug("`charset` 설정 값이 잘못되었습니다.",__LINE__);}
       	if(empty($config['ctype'])){ return $this->debug("`ctype` 설정 값이 잘못되었습니다.",__LINE__);}       	

        // 모듈지정
        if( $config['module'] == 'auto'){
            if( $checksumSmtp === true){  $this->module = 'smtp'; }
            else{ $this->module = 'phpmail'; }
        }
        else{
        	$this->module = $config['module'];
        }        
        return true;
    }

    /*연결종료 */
    private function close(){
        switch($this->module){
            case "smtp":
				if( $this->fp ){
					$this->smtp('dialogue',221, "QUIT");
					fclose($this->fp);
					$this->msg('SMTP close',__LINE__);
				}
            break;

            case "phpmail":
            	$this->msg('PHPMAIL close',__LINE__);
            break;
         }    	
         $this->connection = false;
         return true;
    }


    /*sendmail*/
    function sendmail(){

        $this->msg('PHP VERSION: '.phpversion(),__LINE__); 
        $this->msg('HOST: '.$_SERVER['HTTP_HOST'],__LINE__);
        $this->msg('multiSendMail - sendmail 시작',__LINE__);

    	// 비동기 소켓통신 응답 체크 
    	$asyncCheck = $this->isAsyncSocketClient(); 

        // 설정체크
        $rst = $this->configCheck();
        if($rst === false){ return false; }
        $config = $this->config;
        $option = $this->option;


        /*비동기 소켓통신 전송 체크*/
        if( $config['async.use'] == 1 && $this->isAsyncSocketClient !== true ){

        	if( empty($config['async.url'])){ return $this->debug("`async.url` 설정값이 잘못되었습니다.(통신할 전체 url 을 지정해 주세요)",__LINE__);    }


        	$url = $config['async.url'];
        	$parts = array();
        	foreach($this->parts as $k=>$v){     			       		
        		$parts[$k] = $this->parts[$k]; 
        		unset($parts[$k]['message']);  
        	} // 파일전송시 바이너리 파일 미포함 

        	if(empty($config['async.encrypt'])){ $config['async.encrypt'] = 'md5'; }


        	$hashKey = hash($config['async.encrypt'],$config['async.key']);
        	$datas = array(
	    		'config'=>$config,
	    		'option'=>$option,
	    		'parts'=>$parts,
        	);

        	$params = array($hashKey=>base64_encode(serialize($datas)),'AUTH_ASYNC_SOCKET_CLIENT_KEY'=> $hashKey);
        	$this->applyAsyncSocketClient($url,$params);
        	$this->msg('비동기 소켓통신 - ('.$config['async.url'].') 전송완료 ',__LINE__);
        	return true;
        }

        // 모듈에 따른 처리 
        switch($this->module){

        	// smtp
            case "smtp":
            	return $this->smtp('mailer');
            break;

            // php mail
            case "phpmail":
            	return $this->phpmail('mailer');
            break;
        }
    }


    /*
		- 전체보기
		- 설정/옵션/첨부파일
    */
    function viewSet(){
		$this->viewConfig(true);
		$this->viewOption(true);
		$this->viewAttach(true);
		return true;
    }

    /*
		- 전체삭제 
		- 설정/옵션/첨부파일
    */
    function clearSet(){
    	$this->clearOption(true);
    	$this->clearConfig(true);
    	$this->clearAttach(true);
    	return true;
    }

    /* 옵션세팅 */
    function setOption($key = ''){
        if( empty($key)){return false; }
        if( in_array($key , array_keys($this->option)) < 1){return false; }

        $args = func_get_args();
        if( count($args ) > 0){
            unset($args[0]);
            $args = array_values($args);
        }else{ $args  = array(); }

        $option = array();
        switch($key){

            // 받는이는 Que에 계속 쌓는다.
            case "to":
                if( empty($args[0])){ return false; }
                $option['email'] = $args[0];
                if( !empty($args[1])) $option['name'] = $args[1];
                $this->option[$key][] = $option;
            break;

            // 보는이는 교체방식으로 진행
            case "from":
                if( empty($args[0])){ return false; }
                $option['email'] = $args[0];
                if( !empty($args[1])) $option['name'] = $args[1];
                $this->option[$key] = $option;
            break;

            // 제목
            case "subject":
                if( empty($args[0])){ return false; }
                $option = $args[0];
                $this->option[$key] = $option;
            break;

            // 내용
            case "body":
                if( empty($args[0])){ return false; }
                $option = $args[0];
                $this->option[$key] = $option;
            break;            

            // 참조
            case "cc":
                if( empty($args[0])){ return false; }
                $option['email'] = $args[0];
                if( !empty($args[1])) $option['name'] = $args[1];
                $this->option[$key][] = $option;
            break;   

            // 숨은참조
            case "bcc":
                if( empty($args[0])){ return false; }
                $option['email'] = $args[0];
                if( !empty($args[1])) $option['name'] = $args[1];
                $this->option[$key][] = $option;
            break;                     
        }
    }   

    /*
        - 옵션보기 
        - $key == true 일경우 전체보기
    */
    function viewOption($key = ''){
        $view = array();
        $option = array();
        if( empty($key) || ($key !== true && !isset($this->option[$key]) )  ){
        	return $this->debug("`옵션` 보기에 실패 하였습니다.(설정키값이 존재하지 않습니다.)",__LINE__); 
        }

        if( $key !== true ){
            $option[$key] = $this->option[$key];
        }else{
            $option = $this->option;
        }   

        $view[] = '<ul style="margin:5px 0;padding:0; border:solid 1px #eee; overflow:hidden;">';
        foreach($option as $k=>$v){
            switch($k){
                case "to": // 받는이
                case "cc": // 참조
                case "bcc": // 숨은참조
                     $view[] = '<li style="margin:3px 3px; overflow:hidden; padding:0;list-style-type:none; background-color:#eee;"><strong>['.$k.']</strong>';
                     if( count($option[$k]) < 1){  $view[] = ' <font color="red">NULL</font>'; }
                     else{
                        foreach($option[$k] as $ok=>$ov){

                            $view[] = '<div style="margin:10px 20px; background-color:#fafafa;">';
                            $view[] = '<strong style="color:#999;">[email]</strong> <font>'.$ov['email'].'</font>';
                            if(!empty($ov['name'])) {$view[] = ' <span style="background-color:#fff;display:block;"></span> <strong  style="color:#999;">[name]</strong> <font>'.$ov['name'].'</font>';}
                            $view[] = '</div>';
                        }
                     }
                    $view[] = '</li>';
    
                break;

                // 보내는이
                case "from":
                     $view[] = '<li style="margin:3px 3px; overflow:hidden; padding:0;list-style-type:none; background-color:#eee;"><strong>['.$k.']</strong>';
                    if( empty($v['email'])){  $view[] = ' <font color="red">NULL</font>'; }
                    else{
                        $view[] = '<div style="margin:10px 20px; background-color:#fafafa;">';
                        $view[] = '<strong style="color:#999;">[email]</strong> <font>'.$v['email'].'</font>';
                        if(!empty($v['name'])) {$view[] = ' <span style="background-color:#fff;display:block;"></span> <strong style="color:#999;">[name]</strong> <font>'.$v['name'].'</font>';}
                        $view[] = '</div>';
                    }
                    $view[] = '</li>';
                break;

                
                case "subject": // 제목
                case "body": // 내용
                     $view[] = '<li style="margin:3px 3px; overflow:hidden; padding:0;list-style-type:none; background-color:#eee;"><strong>['.$k.']</strong>';
                    if( empty($v)){  $view[] = ' <font color="red">NULL</font>'; break; }
                    $view[] = ' <font color="blue">'.$v.'</font></li>';
                break;
            }
        }
        $view[] = '</ul>';

        echo '<h3>Option View</h3>'.implode("",$view);
        return true;
    }   


    /*
        - 옵션삭제 
        - $key == true 일경우 전체삭제
    */
    function clearOption($key = ''){
        if( empty($key) || ($key !== true && !isset($this->option[$key]) )  ){
        	return $this->debug("`옵션` 삭제에 실패 하였습니다.(설정키값이 존재하지 않습니다.)",__LINE__); 
        }

        $option = array();
        if( $key !== true){
            $option[$key] = $this->option[$key];
        }
        else{
            $option = $this->option;
        }

        foreach($option as $k=>$v){
            switch($k){
                // 받는이/보내는이/참조/숨은참조
                case "to":
                case "from":
                case "cc":
                case "bcc":
                    $this->option[$k] = array();
                break;

                // 제목/내용
                case "subject":
                case "body":
                    $this->option[$k] = '';
                break;
            }
        }
    }   



    /* 설정세팅 */
    function setConfig($key = '', $value = ''){
        if( empty($key) ){ return false; }
        if( in_array($key , array_keys($this->config)) < 1){ return false; }
        $this->config[$key] = $value;
        return true;
    }

    /*
        - 설정초기화
        - $key == true 일경우 전체초기화
    */
    function clearConfig($key = ''){
        if( empty($key) || ($key !== true && !isset($this->config[$key]) ) ){
        	return $this->debug("`설정` 삭제에 실패 하였습니다.(설정키값이 존재하지 않습니다.)",__LINE__); 
        }

        $config = array();
        if( $key !== true){
            $config[$key] = $this->config[$key];
        }
        else{
            $config = $this->config;
        }
        foreach($config as $k=>$v){
        	$this->config[$k] = $this->clearConfig[$k];
        }
    }

    /*
        - 설정보기
        - $key == true 일경우 전체보기
    */
    function viewConfig($key = '')
    {   
        $view = array();
        $config = array();

        if( empty($key) || ($key !== true && !isset($this->config[$key]) ) ){
        	return $this->debug("`설정` 보기에 실패 하였습니다.(설정키값이 존재하지 않습니다.)",__LINE__); 
        }

        if( $key !== true ){
            $config[$key] = $this->config[$key];
        }else{
            $config = $this->config;
        }   

        $view[] = '<ul style="margin:5px 0;padding:0; border:solid 1px #eee;overflow:hidden;">';
        foreach($config as $k=>$v){
            $view[] = '<li style="margin:3px 3px; padding:0;list-style-type:none;overflow:hidden;"><strong>['.$k.']</strong> '.(empty($v) ? '<font color="red">NULL</font>':'<font color="blue">'.$v.'</font>').'</li>';
        }
        $view[] = '</ul>';

        echo '<h3>Config View</h3>'.implode("",$view);
        return true;
    }


    /* 첨부파일 등록 */
    function setAttach($path, $name="", $ctype="application/octet-stream") {

    	if( empty($path)){ return $this->debug("`첨부파일` 등록에 실패 하였습니다.(파일경로가 없습니다.)",__LINE__);  }
    	
    	$attcnt = count($this->parts); // 해당 값은 삭제 시 키값으로 사용된다.
        if(is_file($path)) {
            $fp = fopen($path, "r");
            $message = fread($fp, filesize($path));
            fclose($fp);
            $this->parts[$attcnt] = array ("ctype" => $ctype, "message" => $message, "name" => $name,"path"=>$path);
        } else{
        	return $this->debug("`첨부파일` 등록에 실패 하였습니다.(`".$path."` 파일이 존재하지 않습니다.)",__LINE__);
        }

        return $attcnt; 
    }

    /*
		- 첨부파일 보기 
		- $key 값이 true 일 경우 전체보기
    */
    function viewAttach($key = '')
    {   
        $view = array();
        $config = array();

    	if( empty($key) || ($key !== true && !isset($this->parts[$key])) ){
    		return $this->debug("`첨부파일` 보기에 실패 하였습니다.(등록된 첨부파일이 없습니다.)",__LINE__); 
    	}

    	$parts = array();
        if( $key !== true ){
            $parts[$key] = $this->parts[$key];
        }else{
            $parts = $this->parts;
        }   

        $view[] = '<ul style="margin:5px 0;padding:0; border:solid 1px #eee;overflow:hidden;">';
        foreach($parts as $k=>$v){
            $view[] = '<li style="margin:3px 3px; padding:0;list-style-type:none;overflow:hidden;"><strong>['.$k.']</strong> '.(empty($v) ? '<font color="red">NULL</font>':'<font color="blue">'.$v['path'].'</font>').'</li>';
        }
        $view[] = '</ul>';

        echo '<h3>Attach View</h3>'.implode("",$view);
        return true;
    }


    /*
        - 첨부파일 비우기
        - $key == true 일경우 전체초기화
    */
    function clearAttach($key = ''){

    	$parts = $this->parts;
    	$attcnt = count($this->parts); 
    	if( empty($key) || ($key !== true && !isset($this->parts[$key])) ){
    		return $this->debug("`첨부파일` 삭제에 실패 하였습니다.(등록된 첨부파일이 없습니다.)",__LINE__); 
    	}
    	if( $key !== true){ unset($this->parts[$key]);}
    	else{ $this->parts = array(); }
    	return true;
    }

    /* 메시지 기록 */
    function msg($msg = '',$line = ''){
        if( empty($msg)){ return false; }

        if( !empty($line)){  $msg = htmlspecialchars('<Line:'.$line.'> ').$msg; }
        $msg =  '['.date('Y-m-d H:i:s').'] '.$msg;

        ob_start();

        if( $this->isAsyncSocketClient === true){
        	echo $msg.$this->newline; 
        }
        else{
        	echo '<div style="background-color:#333; color:#fff; font-size:14px; padding:4px;">'.$msg.'</div>'; 
        }
        
        $viewMsg =  ob_get_clean();
        $this->msg[] = $viewMsg;
        if( $this->config['msg'] == 1){ echo $viewMsg; flush(); }
        return true;
    }

    /* 디버깅 */
    function debug($msg = '',$line = '',$return = false){

        if( empty($line) || empty($msg)){ return $return; }
        if( !empty($line)) { $this->error['line'] = $line; }
        if( !empty($msg)) { $this->error['msg'] = $msg; }
        if($this->config['debug'] == 1){

	        if( $this->isAsyncSocketClient === true){
	        	ob_start();
	        	echo '<pre>';
	        	print_r($this->error);
	        	echo '</pre>';
	        	$debug = ob_get_clean();
	        	echo $debug;
	        }        	
	        else{
	        	$this->support('preview',$this->error);
	        }
        }
        return $return;
    }


    /* 서포트 지원 메소드 */
    function support(){

        $args = func_get_args();
        if( count($args ) > 0){
            $method = $args[0];
            unset($args[0]);
            $args = array_values($args);
            if(!empty($method)) $method = strtolower($method);
            else{ $method = null; }
        }else{ $args  = array(); }
        if( $method === null ){ return false;}

        switch($method){
            // 필수 함수 
            case strtolower("parm"):
                if( count($args[1]) < 1){ return false; }
                $parmSet = array(); $idx = 0;
                foreach($args[1] as $key => $val){ 
                    if( !empty($args[0][$idx])) $parmSet[$key] = $args[0][$idx]; 
                    else $parmSet[$key] = $args[1][$key];
                    $idx++;
                }           
                return $parmSet;
            break;

            // array 뷰
            case strtolower('preview'):
                $parmSet = $this->support('parm',$args,array('data'=>false));
                extract($parmSet);            

                ob_start();
                echo '<pre>';
                print_r($data);
                echo '</pre>';
                $preview= ob_get_clean();
                preg_match_all("/`(.*?)`/is",$preview,$matches);
                if( !empty($matches[1]) && is_array($matches[1]) && count($matches[1]) > 0){
                    foreach($matches[1] as $k=>$v){
                        if( empty($v)) continue;
                        $preview = str_replace($v,'<strong style="color:red">'.$v.'</strong>',$preview);
                        $preview = str_replace(array('<strong style="color:red"><strong style="color:red">','</strong></strong>'),array('<strong style="color:red">','</strong>'),$preview);
                    }
                }

                $arrLineVeiw = explode("\n",$preview);
                unset($arrLineVeiw[0],$arrLineVeiw[1],$arrLineVeiw[count($arrLineVeiw)]);
                $preview = implode("<br>",$arrLineVeiw);
                $preview = '<div style="margin:10px 0; font-size:14px; padding:10px 5px;border:solid 1px #eee;background-color:#fafafa;">'.$preview.'</div>';

                echo $preview;
            break;

            case strtolower('emailChk'):
                $parmSet = $this->support('parm',$args,array('email'=>false));
                extract($parmSet);     

                if( empty($email)){ return false; }
                if(!preg_match("/([\._0-9a-zA-Z-]+)@([0-9a-zA-Z-]+\.[a-zA-Z\.]+)/", $email, $matches)) return false;

                return true;                
            break;


            case strtolower('msgid'):
                $id = date("YmdHis",time());
                mt_srand((float) microtime() * 1000000);
                $randval = mt_rand();
                $id .= $randval."@phpmail";
                return $id;
            break;

            case strtolower('boundary'):
                $uniqchr = uniqid(time());
                $one = strtoupper($uniqchr[0]);
                $two = strtoupper(substr($uniqchr,0,8));
                $three = strtoupper(substr(strrev($uniqchr),0,8));
                return "----=_NextPart_000_000${one}_${two}.${three}";
            break;

        }
    }


}